#
# Copyright (c) 2025 Commonwealth Scientific and Industrial Research Organisation (CSIRO). All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file. See the AUTHORS file for names of contributors.
#
import io
from typing import Any, Optional, Union
from dataclasses import dataclass
from pydantic import ConfigDict, Field, BaseModel, PrivateAttr

from ivcap_client import IVCAP

from .events import EventReporter

class ExecutionContext:
    pass

class JobContext(BaseModel):
    job_id: Optional[str] = None
    report: Optional[EventReporter] = None
    job_authorization: Optional[str] = None

    _ivcap: Optional[IVCAP] = PrivateAttr(None)

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @property
    def ivcap(self) -> Optional[IVCAP]:
        if self._ivcap is None:
            self._ivcap = IVCAP()
        return self._ivcap

@dataclass
class BinaryResult():
    """If the result of the tool is a non json serialisable object, return an
    instance of this class indicating the content-type and the actual
    result either as a byte array or a file handle to a binary content (`open(..., "rb")`)"""
    content_type: str = Field(description="Content type of result serialised")
    content: Union[bytes, str, io.BufferedReader] = Field(description="Content to send, either as byte array or file handle")

@dataclass
class IvcapResult(BinaryResult):
    isError: bool = False
    raw: Any = None

class ExecutionError(BaseModel):
    """
    Pydantic model for execution errors.
    """
    jschema: str = Field("urn:ivcap:schema.ai-tool.error.1", alias="$schema")
    error: str = Field(description="Error message")
    type: str = Field(description="Error type")
    traceback: Optional[str] = Field(None, description="traceback")



#
# Copyright (c) 2025 Commonwealth Scientific and Industrial Research Organisation (CSIRO). All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file. See the AUTHORS file for names of contributors.
#
import json
import os
import traceback
from typing import Callable, Optional, Union, BinaryIO

from time import sleep
from urllib.parse import urlparse, urlunparse
import httpx
from pydantic import BaseModel, HttpUrl
from ag_ui.core.events import BaseEvent

from .events import EventReporter
from .logger import getLogger
from .types import BinaryResult, ExecutionError, IvcapResult

logger = getLogger("ivcap")
event_logger = getLogger("ivcap.event")

# Number of attempt to deliver job result before giving up
MAX_DELIVER_RESULT_ATTEMPTS = 4

OnResultF = Callable[[Union[IvcapResult, ExecutionError], str, Optional[str]], None]

result_callback: OnResultF = None

def set_result_callback(cbk: OnResultF):
    """
    Sets the result handler function to be used for processing job results.

    Args:
        handler: A callable that takes three arguments: the result, job_id,
        and the auth token of the incoming request.
    """
    global result_callback
    result_callback = cbk


def verify_result(result: any, job_id: str, logger) -> any:
    if isinstance(result, ExecutionError):
        return result
    if isinstance(result, BaseModel):
        try:
            return IvcapResult(
                content=result.model_dump_json(by_alias=True),
                content_type="application/json",
                raw=result,
            )
        except Exception as ex:
            msg = f"{job_id}: cannot json serialise pydantic isntance - {str(ex)}"
            logger.warning(msg)
            return ExecutionError(
                error=msg,
                type=type(ex).__name__,
                traceback=traceback.format_exc()
            )
    if isinstance(result, BinaryResult):
        return IvcapResult(content=result.content, content_type=result.content_type)
    if isinstance(result, str):
        return IvcapResult(content=result, content_type="text/plain", raw=result)
    if isinstance(result, bytes):
        # If it's a byte array, return it as is
        return IvcapResult(
            content=result,
            content_type="application/octet-stream",
            raw=result,
        )
    if isinstance(result, BinaryIO):
        # If it's a file handler, return it as is
        return IvcapResult(
            content=result,
            content_type="application/octet-stream",
            raw=result
        )
    # normal model which should be serialisable
    try:
        result = IvcapResult(
            content=json.dumps(result),
            content_type="application/json"
        )
    except Exception as ex:
        msg = f"{job_id}: cannot json serialise result - {str(ex)}"
        logger.warning(msg)
        result = ExecutionError(
            error=msg,
            type=type(ex).__name__,
        )

def push_result(result: Union[IvcapResult, ExecutionError], job_id: str, authorization: Optional[str]=None):
    """Actively push result to sidecar, fail quietly."""
    if result_callback is not None:
        # If a result handler is set, call it as well
        result_callback(result, job_id, authorization)

    ivcap_url = get_ivcap_url()
    if ivcap_url is None:
        if result_callback is None:
            logger.warning(f"{job_id}: no ivcap url found - cannot push result")
        return
    url = urlunparse(ivcap_url._replace(path=f"/results/{job_id}"))

    content_type="text/plain"
    content="SOMETHING WENT WRONG _ PLEASE REPORT THIS ERROR"
    is_error = False
    if not (isinstance(result, ExecutionError) or isinstance(result, IvcapResult)):
        msg = f"{job_id}: expected 'IvcapResult' or 'ExecutionError' but got {type(result)}"
        logger.warning(msg)
        result = ExecutionError(
            error=msg,
            type='InternalError',
        )

    if isinstance(result, IvcapResult):
        content = result.content
        content_type = result.content_type
    else:
        is_error = True
        if not isinstance(result, ExecutionError):
            # this should never happen
            logger.error(f"{job_id}: expected 'ExecutionError' but got {type(result)}")
            result = ExecutionError(
                error="please report unexpected internal error - expected 'ExecutionError' but got {type(result)}",
                type="internal_error",
            )
        content = result.model_dump_json(by_alias=True)
        content_type = "application/json"


    wait_time = 1
    attempt = 0
    headers = {
        "Content-Type": content_type,
        "Is-Error": str(is_error),
    }
    if not (authorization == None or authorization == ""):
        headers["Authorization"] = authorization

    while attempt < MAX_DELIVER_RESULT_ATTEMPTS:
        try:
            response = httpx.post(
                url=url,
                headers=headers,
                data=content,
            )
            response.raise_for_status()
            return
        except Exception as e:
            attempt += 1
            logger.info(f"{job_id}: attempt #{attempt} failed to push result - will try again in {wait_time} sec - {type(e)}: {e}")
            sleep(wait_time)
            wait_time *= 2

    logger.warning(f"{job_id}: giving up pushing result after {attempt} attempts")


class SidecarReporter(EventReporter):
    def __init__(self, job_id: str, job_authorization: str):
        super().__init__(job_id, job_authorization)
        self._ivcap_url = get_ivcap_url()

    def _send(self, event: BaseEvent):
        event_logger.debug(f"{self.job_id}: {event.model_dump_json(exclude_none=True)}")
        if self._ivcap_url is None:
            return

        url = urlunparse(self._ivcap_url._replace(path=f"/events/{self.job_id}"))
        headers = {"Content-Type": "application/json"}
        if not (self.job_authorization == None or self.job_authorization == ""):
            headers["Authorization"] = self.job_authorization
        try:
            m = event.model_dump()
            m["$schema"] = "urn:ag-ui:schema:event.1"
            data = json.dumps(m)
            response = httpx.post(
                url=url,
                headers=headers,
                data=data,
            )
            response.raise_for_status()
        except Exception as e:
            logger.warning(f"could not deliver event - {e}")

def get_ivcap_url() -> HttpUrl:
    """
    Returns the sidecar URL from the request headers.
    """
    base = os.getenv("IVCAP_BASE_URL")
    if base == "" or base is None:
        return None
    return urlparse(base)



#
# Copyright (c) 2025 Commonwealth Scientific and Industrial Research Organisation (CSIRO). All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file. See the AUTHORS file for names of contributors.
#

# Various "patches" to maiontain context between incoming requests
# and calls to external services within a "session"
#
from dataclasses import dataclass
import functools
from logging import Logger
from ivcap_service import getLogger
import os
from typing import Any, Callable, Literal, Optional
from httpx import URL as URLx
from urllib.parse import urlparse

import urllib

from .types import JobContext

ExecContextF = Callable[[], JobContext]

def otel_instrument(
    with_telemetry: Optional[Literal[True]],
    extension: Optional[Callable[[str], None]],
    logger: Logger,
):
    if with_telemetry == False:
        return
    endpoint = os.environ.get('OTEL_EXPORTER_OTLP_ENDPOINT')
    if endpoint == None:
        if with_telemetry == True:
            logger.warning("requested --with-telemetry but exporter is not defined")
        return

    if os.environ.get("PYTHONPATH") == None:
            os.environ["PYTHONPATH"] = ""
    import opentelemetry.instrumentation.auto_instrumentation.sitecustomize # force internal settings
    logger.info(f"instrumenting for endpoint {endpoint}")
    if extension != None:
        extension(endpoint)
    # Also instrumemt
    try:
        from opentelemetry.instrumentation.requests import RequestsInstrumentor
        RequestsInstrumentor().instrument()
    except ImportError:
        pass
    try:
        import httpx # checks if httpx library is even used by this tool
        from opentelemetry.instrumentation.httpx import HTTPXClientInstrumentor
        HTTPXClientInstrumentor().instrument()
    except ImportError:
        pass

def extend_requests(context_f: ExecContextF):
    from requests import Session, PreparedRequest

    logger = getLogger("app.request")

    # Save original function
    wrapped_send = Session.send

    @functools.wraps(wrapped_send)
    def _send(
        self: Session, request: PreparedRequest, **kwargs: Any
    ):
        logger.debug(f"Instrumenting 'requests' request to {request.url}")
        _modify_request(request, context_f, logger)
        # Call original method
        return wrapped_send(self, request, **kwargs)

    # Apply wrapper
    Session.send = _send

def _modify_request(request, context_f: ExecContextF, logger):
    ctxt = context_f()

    headers = request.headers
    url = request.url
    hostname = _get_hostname(url)
    is_local_url = hostname.endswith(".local") or hostname.endswith(".minikube") or hostname.endswith(".ivcap.net")
    if not is_local_url:
        request.url = _wrap_proxy_url(url, headers)
    job_id = ctxt.job_id if ctxt != None else None
    if job_id != None: # OTEL messages won't have a jobID
        headers["Ivcap-Job-Id"] = job_id
    auth = ctxt.job_authorization if ctxt != None else None
    if auth != None and is_local_url:
        logger.debug(f"Adding 'Authorization' header")
        headers["Authorization"] = auth

def _get_hostname(url):
    try:
        if isinstance(url, URLx):
            return url.host
        if isinstance(url, str):
            return urlparse(url).hostname
    except Exception:
        return ""

ivcap_proxy_url = os.getenv('IVCAP_PROXY_URL')

def _wrap_proxy_url(url, headers):
    global ivcap_proxy_url
    if ivcap_proxy_url == None:
        return url

    if isinstance(url, URLx):
        # ensuring that any 'unset' query parameters are not included
        # in the final URL
        query_params = [(k, v) for k, v in url.params.items() if v is not None]
        url2 = URLx(url).copy_with(params=query_params)
        forward_url = str(url2)
        proxy_url = URLx(ivcap_proxy_url)
    if isinstance(url, str):
        forward_url = url # urllib.parse.quote_plus(url)
        proxy_url = ivcap_proxy_url
    headers["Ivcap-Forward-Url"] = forward_url
    return proxy_url

def extend_httpx(context_f: ExecContextF):
    try:
        import httpx
    except ImportError:
        return

    logger = getLogger("app.httpx")

    # Save original function
    wrapped_send = httpx.Client.send
    def _send(self, request, **kwargs):
        logger.debug(f"Instrumenting 'httpx' request to {request.url}")
        _modify_request(request, context_f, logger)
        # Call original method
        return wrapped_send(self, request, **kwargs)
    # Apply wrapper
    httpx.Client.send = _send

    wrapped_asend = httpx.AsyncClient.send
    def _asend(self, request, **kwargs):
        logger.debug(f"Instrumenting 'httpx' async request to {request.url}")
        _modify_request(request, context_f, logger)
        return wrapped_asend(self, request, **kwargs)
    httpx.AsyncClient.send = _asend

def set_context(context_f: ExecContextF):
    extend_requests(context_f)
    extend_httpx(context_f)


#
# Copyright (c) 2025 Commonwealth Scientific and Industrial Research Organisation (CSIRO). All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file. See the AUTHORS file for names of contributors.
#

__version__ = "???"
try:  # Python > 3.8+
    from importlib_metadata import version
    __version__ = version("ivcap_service")
except ImportError:
    try:
        import pkg_resources
        __version__ = pkg_resources.get_distribution('ivcap_service').version
    except Exception:
        pass

def get_version():
    return __version__


#
# Copyright (c) 2025 Commonwealth Scientific and Industrial Research Organisation (CSIRO). All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file. See the AUTHORS file for names of contributors.
#

import inspect
import os
from typing import Any, Callable, Optional, get_type_hints
from typing import (
    Any,
    Callable,
    Optional,
    Tuple,
    Union,
)
from pydantic import BaseModel, Field

from .types import ExecutionContext
from .utils import clean_description, get_input_type

TOOL_SCHEMA = "urn:sd-core:schema.ai-tool.1"

SERVICE_ID_PLACEHOLDER = "#SERVICE_ID#"

class ToolDefinition(BaseModel):
    jschema: str = Field(default=TOOL_SCHEMA, alias="$schema")
    id: str

    model_config = {
        "populate_by_name": True,
    }
    name: str
    service_id: str = Field(alias="service-id")
    description: str
    fn_signature: str
    fn_schema: dict

def print_tool_definition(
    fn: Callable[..., Any],
    *,
    name: Optional[str] = None,
    service_id: Optional[str] = None,
    description: Optional[str] = None,
    id_prefix: str = "urn:sd-core:ai-tool",
):
    td = create_tool_definition(fn, name=name, service_id=service_id, description=description, id_prefix=id_prefix)
    print(td.model_dump_json(indent=2, by_alias=True))

def create_tool_definition(
    fn: Callable[..., Any], *,
    name: Optional[str] = None,
    service_id: Optional[str] = None,
    description: Optional[str] = None,
    id_prefix: str = "urn:sd-core:ai-tool",
) -> ToolDefinition:
    name = os.getenv("IVCAP_SERVICE_NAME", name or fn.__name__)
    name = name.replace(" ", "_").replace("-", "_").lower()

    signature, description = _generate_function_description(fn, name, exclude_types=[ExecutionContext])

    if service_id == None:
        service_id = os.getenv("IVCAP_SERVICE_ID", SERVICE_ID_PLACEHOLDER)

    #fn_sig = inspect.signature(fn)
    input_type, _ = get_input_type(fn)

    td = ToolDefinition(
        id=f"{id_prefix}.{name}",
        name=name,
        service_id=service_id,
        description=clean_description(description),
        fn_signature=signature,
        fn_schema=input_type.model_json_schema(),
    )
    return td

def _generate_function_description(func: Callable, name: Optional[str] = None, exclude_types: Optional[list] = None) -> Tuple[str, str]:
    """Generates a function description with argument descriptions and return values from a function with a Pydantic argument.

    The function description assumes that all fields in the Pydantic argument are provided as "normal" function arguments.
    This is useful for generating documentation for functions that use Pydantic models, as it makes the function signature
    more readable by showing the individual fields of the Pydantic model as separate arguments.

    Args:
        func: The function to generate a description for. The function should have at least one Pydantic model as an argument.
        name: Optional name to override the function's actual name.
        exclude_types: Optional list of types to exclude from the function description. Any subtypes of these types will also be excluded.

    Returns:
        A tuple containing:
        - The function signature as a string.
        - The rest of the function description, including the docstring, argument descriptions, and return type.
    """
    if exclude_types is None:
        exclude_types = []

    def is_excluded_type(param_type):
        """Check if a type should be excluded based on the exclude_types list."""
        if not exclude_types:
            return False

        # Check if the type is in the exclude_types list
        if param_type in exclude_types:
            return True

        # Check if the type is a subclass of any type in the exclude_types list
        for exclude_type in exclude_types:
            # Only check issubclass if both types are classes
            if (isinstance(param_type, type) and isinstance(exclude_type, type) and
                issubclass(param_type, exclude_type)):
                return True

        return False
    signature = inspect.signature(func)
    type_hints = get_type_hints(func)

    # Get the Pydantic model class
    pydantic_model_class = None
    pydantic_param_name = None
    for param_name, param in signature.parameters.items():
        if hasattr(param.annotation, '__mro__') and BaseModel in param.annotation.__mro__:
            pydantic_model_class = param.annotation
            pydantic_param_name = param_name
            break

    if not pydantic_model_class:
        error_msg = "Function does not have a Pydantic model as an argument."
        return ("", error_msg)

    model_schema = pydantic_model_class.schema()

    # Start with the function's docstring
    description = func.__doc__ or ""

    # Create a new function signature with the Pydantic fields as arguments
    function_name = name if name is not None else func.__name__
    new_signature = f"{function_name}("

    # Collect all parameters
    all_params = []

    # Add non-Pydantic parameters first
    for param_name, param in signature.parameters.items():
        if param_name != pydantic_param_name:
            param_type = type_hints.get(param_name, Any)
            # Skip if the type is in the exclude_types list or is a subtype of an excluded type
            if is_excluded_type(param_type) or param_name in exclude_types:
                continue
            param_type_name = param_type.__name__ if hasattr(param_type, '__name__') else str(param_type)
            all_params.append(f"{param_name}: {param_type_name}")

    # Add Pydantic fields
    for field_name, field_info in model_schema['properties'].items():
        # Skip if the field name is "$schema"
        if field_name == "$schema":
            continue
        # Skip if the field name is in the exclude_types list
        if field_name in exclude_types:
            continue
        # Get the correct type from the Pydantic model
        field_type = field_info.get('type', None)

        # If type is not available in the schema, try to get it from the model's __annotations__
        if field_type is None and hasattr(pydantic_model_class, '__annotations__'):
            annotation = pydantic_model_class.__annotations__.get(field_name)
            if annotation is not None:
                if hasattr(annotation, '__origin__') and annotation.__origin__ is Union:
                    # Handle Optional types (Union[type, NoneType])
                    args = annotation.__args__
                    if len(args) == 2 and args[1] is type(None):
                        field_type = args[0].__name__
                        is_optional = True
                    else:
                        field_type = str(annotation).replace('typing.', '')
                else:
                    field_type = annotation.__name__ if hasattr(annotation, '__name__') else str(annotation)
            else:
                field_type = 'Any'
        else:
            # Handle optional fields
            is_optional = False
            if 'default' in field_info or field_name not in model_schema.get('required', []):
                is_optional = True

            # Map JSON schema types to Python types
            if field_type == 'string':
                field_type = 'str'
            elif field_type == 'integer':
                field_type = 'int'
            elif field_type == 'number':
                field_type = 'float'
            elif field_type == 'boolean':
                field_type = 'bool'
            elif field_type == 'array':
                # Handle array types with their item types
                items = field_info.get('items', {})
                item_type = items.get('type', 'Any')
                # Map item type to Python types
                if item_type == 'string':
                    item_type = 'str'
                elif item_type == 'integer':
                    item_type = 'int'
                elif item_type == 'number':
                    item_type = 'float'
                elif item_type == 'boolean':
                    item_type = 'bool'
                field_type = f"array[{item_type}]"
            elif field_type is None:
                field_type = 'Any'

        # Add Optional[] for optional fields
        if is_optional and not field_type.startswith('Optional['):
            field_type = f"Optional[{field_type}]"

        all_params.append(f"{field_name}: {field_type}")

    # Join all parameters with commas
    new_signature += ", ".join(all_params)
    new_signature += ")"

    # Add return type
    return_type = type_hints.get('return', 'unknown')
    if hasattr(return_type, '__name__'):
        return_type_str = return_type.__name__
    else:
        return_type_str = str(return_type)
    new_signature += f" -> {return_type_str}"

    description = f"{new_signature}\n\n{description}\n\nArguments:\n"

    # Add non-Pydantic parameters first with their descriptions
    for param_name, param in signature.parameters.items():
        if param_name == "$schema":
            continue
        if param_name != pydantic_param_name:
            param_type = type_hints.get(param_name, Any)
            # Skip if the type is in the exclude_types list or is a subtype of an excluded type
            if is_excluded_type(param_type) or param_name in exclude_types:
                continue
            param_type_name = param_type.__name__ if hasattr(param_type, '__name__') else str(param_type)
            description += f"  {param_name}: {param_type_name}\n"

    # Add Pydantic fields with their descriptions
    for field_name, field_info in model_schema['properties'].items():
        if field_name == "$schema":
            continue
        # Skip if the field name is in the exclude_types list
        if field_name in exclude_types:
            continue
        # Get the correct type from the Pydantic model
        field_type = field_info.get('type', None)

        # If type is not available in the schema, try to get it from the model's __annotations__
        if field_type is None and hasattr(pydantic_model_class, '__annotations__'):
            annotation = pydantic_model_class.__annotations__.get(field_name)
            if annotation is not None:
                if hasattr(annotation, '__origin__') and annotation.__origin__ is Union:
                    # Handle Optional types (Union[type, NoneType])
                    args = annotation.__args__
                    if len(args) == 2 and args[1] is type(None):
                        field_type = args[0].__name__
                        is_optional = True
                    else:
                        field_type = str(annotation).replace('typing.', '')
                else:
                    field_type = annotation.__name__ if hasattr(annotation, '__name__') else str(annotation)
            else:
                field_type = 'Any'
        else:
            # Handle optional fields
            is_optional = False
            if 'default' in field_info or field_name not in model_schema.get('required', []):
                is_optional = True

            # Map JSON schema types to Python types
            if field_type == 'string':
                field_type = 'str'
            elif field_type == 'integer':
                field_type = 'int'
            elif field_type == 'number':
                field_type = 'float'
            elif field_type == 'boolean':
                field_type = 'bool'
            elif field_type == 'array':
                # Handle array types with their item types
                items = field_info.get('items', {})
                item_type = items.get('type', 'Any')
                # Map item type to Python types
                if item_type == 'string':
                    item_type = 'str'
                elif item_type == 'integer':
                    item_type = 'int'
                elif item_type == 'number':
                    item_type = 'float'
                elif item_type == 'boolean':
                    item_type = 'bool'
                field_type = f"array[{item_type}]"
            elif field_type is None:
                field_type = 'Any'

        # Add Optional[] for optional fields
        if is_optional and not field_type.startswith('Optional['):
            field_type = f"Optional[{field_type}]"

        description += f"  {field_name}: {field_type}"
        if 'description' in field_info:
            description += f" - {field_info['description']}"
        # Add default value information if available
        if 'default' in field_info:
            default_value = field_info['default']
            description += f" (Defaults to {default_value})"
        description += "\n"

    # Add a note about the return type
    description += f"\nReturns: {return_type_str}"

    # Split the description into signature and the rest
    signature_part = new_signature
    rest_part = description.replace(new_signature, "", 1).strip()

    return (signature_part, rest_part)

# def create_tool_definition(
#     fn: Callable[..., Any],
#     name: Optional[str] = None,
#     description: Optional[str] = None
# ) -> ToolDefinition:
#     fn_to_parse = fn
#     name = name or fn_to_parse.__name__
#     description = description or fn_to_parse.__doc__

#     fn_sig = inspect.signature(fn_to_parse)

#     # Handle FieldInfo defaults
#     def r(param):
#         if isinstance(param.default, FieldInfo):
#             return param.replace(default=inspect.Parameter.empty)
#         else:
#             return param
#     fn_sig = fn_sig.replace(parameters=[r(param) for param in fn_sig.parameters.values()])

#     fn_signature = f"{name}{fn_sig}"
#     fn_schema = create_schema_from_function(name, fn_to_parse)
#     return ToolDefinition(
#         id=f"urn:sd-core:ai-tool:{name}",
#         name=name,
#         description=description,
#         fn_signature=fn_signature,
#         fn_schema=fn_schema.model_json_schema(by_alias=False))

# def create_schema_from_function(
#     name: str,
#     func: Union[Callable[..., Any], Callable[..., Awaitable[Any]]],
#     additional_fields: Optional[
#         List[Union[Tuple[str, Type, Any], Tuple[str, Type]]]
#     ] = None,
#     ignore_fields: Optional[List[str]] = None,
# ) -> Type[BaseModel]:
#     """Create schema from function."""
#     fields = {}
#     ignore_fields = ignore_fields or []
#     params = inspect.signature(func).parameters
#     for param_name in params:
#         if param_name in ignore_fields:
#             continue

#         param_type = params[param_name].annotation
#         param_default = params[param_name].default
#         description = None

#         if get_origin(param_type) is typing.Annotated:
#             args = get_args(param_type)
#             param_type = args[0]
#             if isinstance(args[1], str):
#                 description = args[1]

#         if param_type is params[param_name].empty:
#             param_type = Any

#         if param_default is params[param_name].empty:
#             # Required field
#             fields[param_name] = (param_type, FieldInfo(description=description))
#         elif isinstance(param_default, FieldInfo):
#             # Field with pydantic.Field as default value
#             fields[param_name] = (param_type, param_default)
#         else:
#             fields[param_name] = (
#                 param_type,
#                 FieldInfo(default=param_default, description=description),
#             )

#     additional_fields = additional_fields or []
#     for field_info in additional_fields:
#         if len(field_info) == 3:
#             field_info = cast(Tuple[str, Type, Any], field_info)
#             field_name, field_type, field_default = field_info
#             fields[field_name] = (field_type, FieldInfo(default=field_default))
#         elif len(field_info) == 2:
#             # Required field has no default value
#             field_info = cast(Tuple[str, Type], field_info)
#             field_name, field_type = field_info
#             fields[field_name] = (field_type, FieldInfo())
#         else:
#             raise ValueError(
#                 f"Invalid additional field info: {field_info}. "
#                 "Must be a tuple of length 2 or 3."
#             )

#     return create_model(name, **fields)  # type: ignore



#
# Copyright (c) 2025 Commonwealth Scientific and Industrial Research Organisation (CSIRO). All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file. See the AUTHORS file for names of contributors.
#
import json
import logging
from logging.config import dictConfig
import os
from typing import Any

LOGGING_CONFIG={}

def getLogger(name: str) -> logging.Logger:
    return logging.getLogger(name)

def service_log_config():
    return LOGGING_CONFIG

def set_service_log_config(config: Any):
    global LOGGING_CONFIG
    LOGGING_CONFIG = config
    dictConfig(LOGGING_CONFIG)

def logging_init(cfg_path: str=None):
    if not cfg_path:
        script_dir = os.path.dirname(__file__)
        cfg_path = os.path.join(script_dir, "logging.json")

    with open(cfg_path, 'r') as file:
        config = json.load(file)
        set_service_log_config(config)


#
# Copyright (c) 2025 Commonwealth Scientific and Industrial Research Organisation (CSIRO). All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file. See the AUTHORS file for names of contributors.
#

import ast
import os
import sys
from typing import Callable, Dict, List, Any, Optional, Union
from pydantic import BaseModel, Field

from .service import Service, ServiceContact, ServiceLicense
from .utils import clean_description, file_to_json
from .tool_definition import SERVICE_ID_PLACEHOLDER
from .logger import getLogger

SERVICE_SCHEMA = "urn:ivcap:schema.service.2"
BATCH_CONTROLLER_SCHEMA = "urn:ivcap:schema.service.batch.1"
DEF_POLICY = "urn:ivcap:policy:ivcap.open.service"
DEF_RESOURCE_FILE = "resources.json"

IMAGE_PLACEHOLDER = "#DOCKER_IMG#"

logger = getLogger("service-defintion")

class ResourceRequirements(BaseModel):
    cpu: str = Field(default="500m")
    memory: str = Field(default="1Gi")
    ephemeral_storage: Optional[str] = None # Field(default="1Gi")

class Resources(BaseModel):
    limits: ResourceRequirements = Field(default_factory=ResourceRequirements)
    requests: ResourceRequirements = Field(default_factory=ResourceRequirements)

class BatchController(BaseModel):
    jschema: str = Field(default=BATCH_CONTROLLER_SCHEMA, alias="$schema")
    image: str
    command: Union[List[str], str]
    resources: Resources = Field(default_factory=Resources)

class ServiceDefinition(BaseModel):
    jschema: str = Field(default=SERVICE_SCHEMA, alias="$schema")
    id: str = Field(alias="$id")
    name: str
    description: str
    parameters: List[Any] = []
    contact: ServiceContact = Field(description="contact details of the service")
    license: Optional[ServiceLicense] = Field(None, description="license details of the service")
    policy: str = Field(default=DEF_POLICY)
    controller_schema: str = Field(description="URN identifying the controller for this service", alias="controller-schema")
    controller: Any

    model_config = {
        "populate_by_name": True,
    }

def print_batch_service_definition(
    service_description: Service,
    fn: Callable[..., Any],
    service_id: Optional[str] = None,
):
    sd = create_batch_service_definition(
        service_description,
        fn,
        service_id=service_id,
    )
    print(sd.model_dump_json(indent=2, by_alias=True, exclude_none=True))

def create_batch_service_definition(
    service_description: Service,
    fn: Callable[..., Any],
    service_id: Optional[str] = None,
) -> ServiceDefinition:
    # controller
    image = os.getenv("DOCKER_IMG", IMAGE_PLACEHOLDER)

    command = find_command()
    resources = find_resources_file()
    controller = BatchController(image=image, command=command, resources=resources)
    return create_service_definition(service_description, fn, BATCH_CONTROLLER_SCHEMA, controller, service_id)

def create_service_definition(
    service_description: Service,
    fn: Callable[..., Any],
    controller_schema: str,
    controller: Any,
    service_id: Optional[str] = None,
    description: Optional[str] = None,
) -> ServiceDefinition:
    name = os.getenv("IVCAP_SERVICE_NAME", service_description.name)
    if service_id == None:
        service_id = os.getenv("IVCAP_SERVICE_ID", SERVICE_ID_PLACEHOLDER)
    description = clean_description(fn.__doc__ or "no description available")
    contact = service_description.contact
    license = service_description.license
    policy = os.getenv("IVCAP_POLICY_URN", DEF_POLICY)

    sd_data = {
        "$id": service_id,
        "name": name,
        "description": description,
        "policy": policy,
        "contact": contact,
        "license": license,
        "controller_schema": controller_schema,
        "controller": controller,
    }

    sd = ServiceDefinition(**sd_data)
    return sd

def find_resources_file() -> Resources:
    using_def_resource_file = False
    resource_file = os.getenv("IVCAP_RESOURCES_FILE")
    if resource_file == None:
        using_def_resource_file = True
        resource_file = DEF_RESOURCE_FILE

    if os.path.exists(resource_file) and os.access(resource_file, os.R_OK):
        rd = file_to_json(resource_file)
        resources = Resources(**rd)
    else:
        if using_def_resource_file:
            # should not use logger as this is likely piped into some other command
            print(f"WARNING: Using default resources as I can't find resources def '{resource_file}'", file=sys.stderr)
            resources = Resources()
        else:
            print(f"FATAL: Cannot open resources definition file '{resource_file}'", file=sys.stderr)
            sys.exit(-1)
    return resources

def find_command() -> List[str]:
    docker_file = os.getenv("DOCKERFILE", "Dockerfile")
    if not(os.path.exists(docker_file) and os.access(docker_file, os.R_OK)):
        print(f"FATAL: Cannot find Dockerfile '{docker_file}'", file=sys.stderr)
        sys.exit(-1)

    entry = extract_line(docker_file, "ENTRYPOINT")
    if not entry:
        print(f"FATAL: Cannot find 'ENTRYPOINT' in '{docker_file}'", file=sys.stderr)
        sys.exit(-1)
    cs = entry[len("ENTRYPOINT"):].strip()
    cmd = ast.literal_eval(cs)
    return cmd

def extract_line(filepath, start_string):
    with open(filepath, 'r') as f:
        for line in f:
            if line.startswith(start_string):
                return line.strip()
    return None


#
# Copyright (c) 2025 Commonwealth Scientific and Industrial Research Organisation (CSIRO). All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file. See the AUTHORS file for names of contributors.
#
from contextlib import contextmanager
import time
import traceback
from typing import Any, Callable, Generator, List, Optional
from ag_ui.core.events import (
    TextMessageStartEvent, TextMessageContentEvent, TextMessageEndEvent,
    ToolCallStartEvent, ToolCallArgsEvent, ToolCallEndEvent,
    StateSnapshotEvent, StateDeltaEvent, MessagesSnapshotEvent,
    RawEvent, CustomEvent, RunStartedEvent, RunFinishedEvent,
    RunErrorEvent, StepStartedEvent, StepFinishedEvent, EventType, BaseEvent
)
from ag_ui.core.types import Message, State


from .logger import getLogger

logger = getLogger("event")

event_reporter_factory = None

EventFactoryF = Callable[[str, Optional[str]], 'EventReporter']
def set_event_reporter_factory(factory: EventFactoryF):
    """
    Set afactory function for creating specialised EventReporter instances.

    Args:
        factory (EventFactoryF): A factory function that takes a job ID and an optional job authorization token,
        and returns an instance of EventReporter. If None, the default EventReporter will be used.
    """
    global event_reporter_factory
    if factory is not None and not callable(factory):
        raise ValueError("Factory must be a callable that returns an EventReporter instance.")
    event_reporter_factory = factory

def create_event_reporter(job_id: str, job_authorization: Optional[str] = None) -> 'EventReporter':
    """
    Create an EventReporter instance for the given job ID.


    Args:
        job_id (str): The unique identifier for the job.
        job_authorization (Optional[str]): Optional authorization token for the job.

    Returns:
        EventReporter: An instance of EventReporter initialized with the job ID.
    """
    if event_reporter_factory is not None:
        return event_reporter_factory(job_id, job_authorization)
    return EventReporter(job_id, job_authorization)

class EventContext:
    def __init__(self, event_name:str):
        self._event_name = event_name
        self._exit_msg = None
        self._exit_timestamp = None

    def finished(self, msg: Optional[Any] = None, timestamp: Optional[int] = None):
        self._exit_msg = msg
        self._exit_timestamp = timestamp

    def _finished(self, finish_fn):
        finish_fn(self._event_name, self._exit_msg, self._exit_timestamp)

EventCtxtGenerator = Generator[EventContext, None, None]

class EventReporter:
    def __init__(self, job_id: str, job_authorization: str):
        self.job_id = job_id
        self.job_authorization = job_authorization

    def _send(self, event: BaseEvent):
        logger.debug(f"{self.job_id}: {event.model_dump_json(exclude_none=True)}")

    def _emit(self, cls, event_type, **kwargs):
        try:
            event = cls(type=event_type, **kwargs)
            if event.timestamp is None:
                    event.timestamp = int(time.time() * 1000)
            self._send(event)
        except Exception as e:
            logger.error(f"Failed to emit event {event_type}: {e}")

    def text_message_start(self, message_id: str, role: str, raw_event: Optional[Any] = None, timestamp: Optional[int] = None):
        self._emit(TextMessageStartEvent, EventType.TEXT_MESSAGE_START, message_id=message_id, role=role, raw_event=raw_event, timestamp=timestamp)

    def text_message_content(self, message_id: str, delta: str, raw_event: Optional[Any] = None, timestamp: Optional[int] = None):
        self._emit(TextMessageContentEvent, EventType.TEXT_MESSAGE_CONTENT, message_id=message_id, delta=delta, raw_event=raw_event, timestamp=timestamp)

    def text_message_end(self, message_id: str, raw_event: Optional[Any] = None, timestamp: Optional[int] = None):
        self._emit(TextMessageEndEvent, EventType.TEXT_MESSAGE_END, message_id=message_id, raw_event=raw_event, timestamp=timestamp)

    def tool_call_start(self, tool_call_id: str, tool_call_name: str, parent_message_id: Optional[str] = None, raw_event: Optional[Any] = None, timestamp: Optional[int] = None):
        self._emit(ToolCallStartEvent, EventType.TOOL_CALL_START, tool_call_id=tool_call_id, tool_call_name=tool_call_name, parent_message_id=parent_message_id, raw_event=raw_event, timestamp=timestamp)

    def tool_call_args(self, tool_call_id: str, delta: str, raw_event: Optional[Any] = None, timestamp: Optional[int] = None):
        self._emit(ToolCallArgsEvent, EventType.TOOL_CALL_ARGS, tool_call_id=tool_call_id, delta=delta, raw_event=raw_event, timestamp=timestamp)

    def tool_call_end(self, tool_call_id: str, raw_event: Optional[Any] = None, timestamp: Optional[int] = None):
        self._emit(ToolCallEndEvent, EventType.TOOL_CALL_END, tool_call_id=tool_call_id, raw_event=raw_event, timestamp=timestamp)

    def state_snapshot(self, snapshot: State, raw_event: Optional[Any] = None, timestamp: Optional[int] = None):
        self._emit(StateSnapshotEvent, EventType.STATE_SNAPSHOT, snapshot=snapshot, raw_event=raw_event, timestamp=timestamp)

    def state_delta(self, delta: List[Any], raw_event: Optional[Any] = None, timestamp: Optional[int] = None):
        self._emit(StateDeltaEvent, EventType.STATE_DELTA, delta=delta, raw_event=raw_event, timestamp=timestamp)

    def messages_snapshot(self, messages: List[Message], raw_event: Optional[Any] = None, timestamp: Optional[int] = None):
        self._emit(MessagesSnapshotEvent, EventType.MESSAGES_SNAPSHOT, messages=messages, raw_event=raw_event, timestamp=timestamp)

    def raw(self, event: Any, source: Optional[str] = None, raw_event: Optional[Any] = None, timestamp: Optional[int] = None):
        self._emit(RawEvent, EventType.RAW, event=event, source=source, raw_event=raw_event, timestamp=timestamp)

    def custom(self, name: str, value: Any, raw_event: Optional[Any] = None, timestamp: Optional[int] = None):
        self._emit(CustomEvent, EventType.CUSTOM, name=name, value=value, raw_event=raw_event, timestamp=timestamp)

    def run_started(self, thread_id: str, run_id: str, raw_event: Optional[Any] = None, timestamp: Optional[int] = None):
        self._emit(RunStartedEvent, EventType.RUN_STARTED, thread_id=thread_id, run_id=run_id, raw_event=raw_event, timestamp=timestamp)

    def run_finished(self, thread_id: str, run_id: str, raw_event: Optional[Any] = None, timestamp: Optional[int] = None):
        self._emit(RunFinishedEvent, EventType.RUN_FINISHED, thread_id=thread_id, run_id=run_id, raw_event=raw_event, timestamp=timestamp)

    def run_error(self, message: str, code: Optional[str] = None, raw_event: Optional[Any] = None, timestamp: Optional[int] = None):
        self._emit(RunErrorEvent, EventType.RUN_ERROR, message=message, code=code, raw_event=raw_event, timestamp=timestamp)

    def step_started(self, step_name: str, raw_event: Optional[Any] = None, timestamp: Optional[int] = None):
        self._emit(StepStartedEvent, EventType.STEP_STARTED, step_name=step_name, raw_event=raw_event, timestamp=timestamp)

    def step_finished(self, step_name: str, raw_event: Optional[Any] = None, timestamp: Optional[int] = None):
        self._emit(StepFinishedEvent, EventType.STEP_FINISHED, step_name=step_name, raw_event=raw_event, timestamp=timestamp)

    def step(self, step_name, raw_event=None, timestamp=None):
        return self._event_scope(step_name, self.step_started, self.step_finished, raw_event, timestamp)

    @contextmanager
    def _event_scope(
        self,
        event_name: str,
        start_fn: Callable[[str, Optional[Any], Optional[int]], None],
        finish_fn: Callable[[str, Optional[Any], Optional[int]], None],
        raw_event: Optional[Any] = None,
        timestamp: Optional[int] = None
    ) -> EventCtxtGenerator:
        ctxt = EventContext(event_name)
        start_fn(event_name, raw_event, timestamp)
        try:
            yield ctxt
        except Exception as e:
            import traceback
            trace = traceback.format_exc()
            self.run_error(f"exception in event '{event_name}' - {str(e)}", trace)
            raise e
        ctxt._finished(finish_fn)


#
# Copyright (c) 2025 Commonwealth Scientific and Industrial Research Organisation (CSIRO). All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file. See the AUTHORS file for names of contributors.
#

import time
import traceback
from typing import Callable, Tuple
import argparse
from logging import Logger
from typing import Any, Callable, Dict, Optional
from urllib.parse import urlunparse

import os
import sys
import httpx
from pydantic import BaseModel, ConfigDict, Field
import requests

from ivcap_client import IVCAP

from .context import JobContext, otel_instrument, set_context
from .ivcap import SidecarReporter, get_ivcap_url, push_result, verify_result
from .logger import getLogger
from .types import ExecutionError
from .utils import get_function_return_type, get_input_type
from .version import get_version
from .tool_definition import print_tool_definition  # Import the requests library
from .events import EventReporter, create_event_reporter, set_event_reporter_factory

class ServiceContact(BaseModel):
    name: str = Field(description="name of the contact person")
    email: str = Field(description="email address of the contact person")
    url: Optional[str] = Field(None, description="url of the contact person")

class ServiceLicense(BaseModel):
    name: str = Field(description="name of the license")
    url: str = Field(description="url to the license text")

class Service(BaseModel):
    name: str = Field(description="name of the service")
    version: Optional[str] = Field(os.environ.get("VERSION", "???"), description="version of the service")
    contact: ServiceContact = Field(description="contact details of the service")
    license: Optional[ServiceLicense] = Field(None, description="license of the service")

# Number of attempt to request a new job before giving up
MAX_REQUEST_JOB_ATTEMPTS = 4

class ServiceContext(BaseModel):
    worker_fn: Callable
    input_model: type[BaseModel]
    output_model: type[BaseModel]
    fn_add_job_context: Optional[str]
    job_context: Optional[JobContext] = None
    logger: Logger

    model_config = ConfigDict(arbitrary_types_allowed=True)

def wait_for_work(svc_ctxt: ServiceContext):
    ivcap_url = get_ivcap_url()
    if ivcap_url is None:
        svc_ctxt.logger.warning(f"no ivcap url found - cannot request work")
        return
    url = urlunparse(ivcap_url._replace(path=f"/next_job"))
    logger = svc_ctxt.logger
    logger.info(f"... checking for work at '{url}'")
    try:

        while True:
            result = None
            try:
                (response, job_authorization) = fetch_job(url, logger)
                job = response.json()
                schema = job.get("$schema", "")
                if schema.startswith("urn:ivcap:schema.service.batch.done"):
                    logger.info("no more jobs - we are done")
                    sys.exit(0)

                job_id = job.get("id", "unknown_job_id")  # Provide a default value if "id" is missing
                result = do_job(job, svc_ctxt, job_authorization)
                result = verify_result(result, job_id, logger)
            except Exception as e:
                result = ExecutionError(
                    error=str(e),
                    type=type(e).__name__,
                    traceback=traceback.format_exc()
                )
                logger.warning(f"job {job_id} failed - {result.error}")
            finally:
                if result is not None:
                    logger.info(f"job {job_id} finished, sending result message")
                    push_result(result, job_id, None)

    except requests.exceptions.RequestException as e:
        logger.warning(f"Error during request: {e}")
    except Exception as e:
        logger.warning(f"Error processing job: {e}")

def fetch_job(url: str, logger: Logger) -> Tuple[Any, Optional[str]]:
    wait_time = 1
    attempt = 0
    while attempt < MAX_REQUEST_JOB_ATTEMPTS:
        try:
            response = httpx.get(url)
            response.raise_for_status()
            job_authorization = response.headers.get("authorization")
            return (response, job_authorization)
        except Exception as e:
            attempt += 1
            logger.info(f"attempt #{attempt} failed to fetch new job - will try again in {wait_time} sec - {type(e)}: {e}")
            time.sleep(wait_time)
            wait_time *= 2
    logger.info("cannot contact sidecar - bailing out")
    sys.exit(255)

def do_job(
    job: Any,
    svc_ctxt: ServiceContext,
    job_authorization: Optional[str] = None
):
    job_id = job.get("id", "unknown_job_id")  # Provide a default value if "id" is missing
    ct = job["in-content-type"]
    if ct != "application/json":
        raise Exception(f"cannot handle content-type '{ct}'")
    jc = job["in-content"]
    logger = svc_ctxt.logger
    mreq = svc_ctxt.input_model(**jc)
    logger.info(f"{job_id}: calling worker with - {mreq}")
    reporter = create_event_reporter(job_id=job_id, job_authorization=job_authorization)
    svc_ctxt.job_context = JobContext(job_id=job_id, job_authorization=job_authorization, report=reporter)
    try:
        f = svc_ctxt.worker_fn
        if svc_ctxt.fn_add_job_context is None:
            resp = f(mreq)
        else:
            d = {}
            d[svc_ctxt.fn_add_job_context] = svc_ctxt.job_context
            resp = f(mreq, **d)
        logger.info(f"{job_id}: worker finished with - {resp}")
        if type(resp) != svc_ctxt.output_model:
            logger.warning(f"{job_id}: result is of type '{type(resp)}' but expected '{svc_ctxt.output_model}'")

    except BaseException as ex:
        logger.warning(f"{job_id}: failed - '{ex}'")
        resp = ExecutionError(
                        error=str(ex),
                        type=type(ex).__name__,
                        traceback=traceback.format_exc()
                    )
    return resp

def start_batch_service(
    service_description: Service,
    worker_fn: Callable,
    *,
    custom_args: Optional[Callable[[argparse.ArgumentParser], argparse.Namespace]] = None,
    run_opts: Optional[Dict[str, Any]] = None,
    with_telemetry: Optional[bool] = None,
):
    """A helper function to start a batch service

    Args:
        service_description (Service): description of the service
        tool_fn (Callable[..., Any]): _description_
        logger (Logger): _description_
        custom_args (Optional[Callable[[argparse.ArgumentParser], argparse.Namespace]], optional): _description_. Defaults to None.
        run_opts (Optional[Dict[str, Any]], optional): _description_. Defaults to None.
        with_telemetry: (Optional[bool]): Instantiate or block use of OpenTelemetry tracing
    """
    logger = getLogger("server")

    parser = argparse.ArgumentParser(description=service_description.name)
    parser.add_argument('--with-telemetry', action="store_true", help='Initialise OpenTelemetry')
    parser.add_argument('--print-service-description', action="store_true", help='Print service description to stdout')
    parser.add_argument('--print-tool-description', action="store_true", help='Print tool description to stdout')
    parser.add_argument('--test-file', type=str, help='path to job file for testing service')

    if custom_args is not None:
        args = custom_args(parser)
    else:
        args = parser.parse_args()

    if args.print_service_description:
        from .service_definition import print_batch_service_definition
        print_batch_service_definition(service_description, worker_fn)
        sys.exit(0)

    if args.print_tool_description:
        print_tool_definition(worker_fn, name = service_description.name)
        sys.exit(0)

    logger.info(f"{service_description.name} - {os.getenv('VERSION')} - v{get_version()}")

    set_event_reporter_factory(SidecarReporter)
    svc_ctxt = create_service_context(worker_fn, logger)
    if args.test_file is not None:
        from .utils import file_to_json
        job = file_to_json(args.test_file)
        res = do_job(job, svc_ctxt)
        if get_ivcap_url() is not None:
            result = verify_result(res, job['id'], logger)
            push_result(result, job['id'], None)

        print(res.model_dump_json(indent=2, by_alias=True))
    else:
        otel_instrument(with_telemetry, None, logger)
        set_context(lambda: svc_ctxt.job_context)
        wait_for_work(svc_ctxt)


def create_service_context(worker_fn: Callable, logger: Logger) -> ServiceContext:
    input_model, extras = get_input_type(worker_fn)
    if len(extras) > 1:
        logger.warning(f"worker function '{worker_fn.__name__}' has more than one extra paramter - only JobContext is allowed")
        sys.exit(1)
    fn_job_context_p = None
    if len(extras) == 1:
        fn_job_context_p = list(extras.keys())[0]
        if extras[fn_job_context_p] != JobContext:
            logger.warning(f"worker function '{worker_fn.__name__}' can only have 'JobContext' as additional parameter")
            sys.exit(1)

    output_model = get_function_return_type(worker_fn)

    return ServiceContext(
        worker_fn=worker_fn,
        fn_add_job_context=fn_job_context_p,
        input_model=input_model,
        output_model=output_model,
        logger=logger,
    )



#
# Copyright (c) 2025 Commonwealth Scientific and Industrial Research Organisation (CSIRO). All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file. See the AUTHORS file for names of contributors.
#
import os
from typing import Optional, Dict
import httpx

def file_to_http_response(
    file_path: str,
    headers: Optional[Dict[str, str]] = None,
    status_code: int = 200,
) -> httpx.Response:
    """
    Reads data from a file and returns it as an httpx.Response object.

    Args:
        file_path: The path to the file to read.
        headers: Optional dictionary of HTTP headers to include in the response.
        status_code: The HTTP status code to return (e.g., 200, 404, 500).
        status_message: Optional HTTP status message. If None, the default
            message for the status_code is used.

    Returns:
        An httpx.Response object containing the file data and response information.
    """
    # Sanity checks
    if not os.path.exists(file_path):
        raise FileNotFoundError(f"File not found: {file_path}")
    if not os.path.isfile(file_path):
        raise ValueError(f"Not a file: {file_path}")
    if status_code < 100 or status_code > 599:
        raise ValueError(f"Invalid HTTP status code: {status_code}")

    # Default headers
    default_headers = {
        "Content-Type": "application/octet-stream",  # Default content type
        "Content-Length": str(os.path.getsize(file_path)),
    }

    # Combine default and user-provided headers
    combined_headers = default_headers.copy()
    if headers:
        combined_headers.update(headers) # Use update for simpler merging

    # Read the file data
    try:
        with open(file_path, "rb") as file:
            file_data = file.read()
    except Exception as e:
        raise IOError(f"Error reading file: {file_path}: {e}")

    # Create an httpx.Response object
    response = httpx.Response(
        status_code=status_code,
        content=file_data,
        headers=combined_headers,
    )

    return response



#
# Copyright (c) 2025 Commonwealth Scientific and Industrial Research Organisation (CSIRO). All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file. See the AUTHORS file for names of contributors.
#
""" A library for building services for the IVCAP platform"""

from .version import __version__
from .logger import getLogger, logging_init, service_log_config, set_service_log_config
from .service import start_batch_service, Service
from .service_definition import create_service_definition, find_resources_file, find_command, IMAGE_PLACEHOLDER, Resources, ServiceDefinition
from .tool_definition import create_tool_definition, print_tool_definition, ToolDefinition
from .utils import get_function_return_type, get_input_type
from .ivcap import get_ivcap_url, verify_result, push_result, set_result_callback, OnResultF, SidecarReporter
from .types import IvcapResult, BinaryResult, ExecutionError, JobContext
from .context import otel_instrument, set_context
from .events import EventReporter, set_event_reporter_factory, EventFactoryF, create_event_reporter


#
# Copyright (c) 2025 Commonwealth Scientific and Industrial Research Organisation (CSIRO). All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file. See the AUTHORS file for names of contributors.
#
import inspect
import json
import os
from typing import Optional, Type, Callable, Any, get_type_hints, Dict, Tuple

from pydantic import BaseModel

def get_input_type(func: Callable) -> Tuple[Optional[Type[BaseModel]], Dict[str, Any]]:
    """Gets the input type of a function.

    Args:
        func: The function to get the input type for.

    Returns:
        A tuple containing:
        - The first function parameter which is a derived class of a pydantic BaseModel, or None if no such parameter exists.
        - A dictionary of all additional parameters, where the key is the parameter name and the value is the type.
    """
    signature = inspect.signature(func)
    type_hints = get_type_hints(func)

    # Get the Pydantic model class
    pydantic_model_class = None
    pydantic_param_name = None
    for param_name, param in signature.parameters.items():
        if hasattr(param.annotation, '__mro__') and BaseModel in param.annotation.__mro__:
            pydantic_model_class = param.annotation
            pydantic_param_name = param_name
            break

    # Get all additional parameters
    additional_params = {}
    for param_name, param in signature.parameters.items():
        if param_name != pydantic_param_name:
            param_type = type_hints.get(param_name, Any)
            additional_params[param_name] = param_type

    return pydantic_model_class, additional_params

def get_function_return_type(func):
    """Extracts the return type from a function."""
    type_hints = get_type_hints(func)
    # param_types = {k: v for k, v in type_hints.items() if k != 'return'}
    return_type = type_hints.get('return')
    # return param_types, return_type
    return return_type

def file_to_json(file_path: str) -> Dict[str, Any]:
    """
    Reads a file, attempts to parse it as JSON, and returns the content.

    Args:
        file_path: The path to the file to read.

    Returns:
        A dictionary containing the parsed JSON content.

    Raises:
        FileNotFoundError: If the file does not exist.
        ValueError: If the file is not a valid JSON file.
        IOError: If there is an error reading the file.
    """
    if not os.path.exists(file_path):
        raise FileNotFoundError(f"File not found: {file_path}")

    try:
        with open(file_path, "r") as f:
            return json.load(f)
    except FileNotFoundError as e:
        raise FileNotFoundError(f"Error finding file: {file_path}: {e}")
    except json.JSONDecodeError as e:
        raise ValueError(f"Error decoding JSON from file: {file_path}: {e}")
    except Exception as e:
        raise IOError(f"Error reading file: {file_path}: {e}")

def clean_description(text: str) -> str:
    """
    Cleans a string by removing unnecessary whitespace and newlines, while preserving double newlines.
    """
    text = text.replace("\n\n", "TEMP_NEWLINE")
    text = text.replace("\n", " ")
    text = " ".join(text.split())
    text = text.replace("TEMP_NEWLINE", "\n")
    return text
